/* %Z%%M% %I% %E% 1989 by Robin T. Miller. */

/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1989 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			       14 Aladdin Road 				    *
 *			      Windham, NH 03087				    *
 *			       (603) 890-8723				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Modification History:
 *
 * November 22, 1995 by Robin Miller.
 *	Initial port to QNX 4.21 Operating System.
 */

#include "tape_dup.h"
#include <malloc.h>
#if defined(_QNX_SOURCE)
#  include <sys/qioctl.h>
#else /* !defined(_QNX_SOURCE) */
#  include <sys/ioctl.h>
#  include <sys/mtio.h>
#endif /* defined(_QNX_SOURCE) */

/************************************************************************
 *									*
 * tape_write() - Write files to tape.					*
 *									*
 * Description:	This function is used to write files previously read	*
 *		from tape onto disk back onto the tape.  An information	*
 *		file originally created when the file was created is	*
 *		used to verify the data integrity via crc, the block	*
 *		size of each record, and the total file size.		*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE.				*
 *									*
 ************************************************************************/
int
tape_write(void)
{
	char tbuff[DIR_SIZE];
	char *bufp = tbuff;
	u_short crc = 0;
	u_int info_crc;
	ssize_t read_count, write_count;
	int size, info_record, info_size, status = SUCCESS;
	int end_of_file = TRUE;
	int end_of_tape = FALSE;

	/*
	 * Find the total number of files to write to tape.
	 */
	(void) sprintf (bufp, "%s/%s", directory_name, FILE_TOTALS);
	if ( (fpi = fopen (bufp, "r")) == NULL) {
	    fprintf (stderr, "%s: Unable to open totals file %s\n",
							name, bufp);
	    return (FAILURE);
	}
	(void) fscanf (fpi, "%d\n", &total_files);
	if (verbose_flag) {
	    printf ("A total of %d files will be %s.\n", total_files,
			(verify_flag) ? "verified" : "written to the tape");
	}
	(void) fclose (fpi);
	/*
	 * Loop reading all or specified files to/from the tape.
	 */
	for (;;) {
	    if (end_of_file == TRUE) {
		end_of_file = FALSE;
	        if ((status = tw_open_files()) != SUCCESS) {
		    return (status);
		}
	    }
	    /*
	     * The info file tells us how many bytes to read and then
	     * write to the output file. The info file format is:
	     *
	     *	record_number block_size crc_bytes \n
	     *       .
	     *       .
	     *	-1 total_file_size 0
	     *
	     */
	    if ((fscanf (fpi, "%d %d %x\n", &info_record, &info_size,
					&info_crc)) == EOF) {
		fprintf (stderr,
	"%s: Premature end of file detected reading info file.\n", name);
		return (FAILURE);
	    }
	    /*
	     * If the record count is minus, this is the last record.
	     */
	    if (info_record < 0) {
		if (info_size != file_size) {
		    fprintf (stderr,
	"%s: The file size doesn't match size from the info file.\n", name);
		    fprintf (stderr,
	"%s: File %d: Expected file size is %d, current file size is %d\n",
				name, file_number, info_size, file_size);
		    return (FAILURE);
		}
		if ((status = tw_proc_eof()) != SUCCESS) {
		    return (status);
		}
		file_number++;
		record_count = 0;
		starting_record = 0;
		total_size += file_size;
		file_size = 0;
		end_of_file = TRUE;
		(void) fclose (fp);
		(void) fclose (fpi);
		if ( (file_number > file_count) || (file_number > total_files) ) {
		    --file_number;
		    break;
		}
		continue;
	    }
	    record_count++;
	    if (info_record != record_count) {	/* Sanity check. */
		fprintf (stderr,
	"%s: Info record count doesn't match current record count.\n", name);
		fprintf (stderr, "%s: Info record = %d, current record = %d\n",
					name, info_record, record_count);
		return (FAILURE);
	    }
	    /*
	     * If the allocated buffer isn't large enough, reallocate it.
	     */
	    if (block_size < info_size) {
		if (debug_flag) {
		   printf ("Allocating larger buffer, old size = %d, new size = %d\n",
						block_size, info_size);
		}
		data_buffer = (iobuf_t) realloc (data_buffer, info_size);
		block_size = info_size;
#if defined(_QNX_SOURCE)
		if (setvbuf (fp, (char *)data_buffer, _IOFBF, block_size) < 0) {
		    perror("setvbuf()");
		}
#else /* !defined(_QNX_SOURCE) */
		setbuffer (fp, (char *)data_buffer, block_size);
#endif /* defined(_QNX_SOURCE) */
	    }
	    if (debug_flag) {
		printf ("File %d: Reading %d bytes...\n",
						file_number, info_size);
	    }
	    if ((read_count = fread (data_buffer, sizeof(char), info_size, fp)) != info_size) {
		fprintf (stderr,
		"%s: Error reading data record, count = %d, expected = %d\n",
					name, read_count, info_size);
		return (FAILURE);
	    }
	    /*
	     * Verify the crc against the original from the info file.
	     */
	    crc = 0;
	    file_size += read_count;
	    crc = updcrc (crc, data_buffer, read_count);
	    if (crc != info_crc) {
		fprintf (stderr,
	"%s: CRC does not match original CRC, crc = %x, expected = %x\n",
				name, crc, info_crc);
		return (FAILURE);
	    }
	    if (verify_flag == FALSE) {
		if ((write_count = write (ofd, data_buffer, read_count)) != read_count) {
		    fprintf (stderr, "%s: write (%d) != read (%d)\n",
					name, write_count, read_count);
		    return (EIO);
		}
	    }
	}
	(void) close (ofd);		/* Close writes last file mark.	*/
	file_size = 0;			/* Complete file was processed.	*/
	return (SUCCESS);
}

/************************************************************************
 *									*
 * tw_open_files() - Open disk files.					*
 *									*
 * Description:	This function is used to open the disk data and info	*
 *		files to be written to the tape.			*
 *									*
 * Implicit Inputs:							*
 *		file_number - Used to construct the file names to open.	*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE.				*
 *									*
 ************************************************************************/
int
tw_open_files()
{
	char tbuff[DIR_SIZE];
	char *bufp = tbuff;

	(void) sprintf (bufp, "%s/file_%d", directory_name, file_number);
	if ( (fp = fopen (bufp, "r")) == NULL) {
	    fprintf (stderr, "%s: Unable to open input file %s\n",
							name, bufp);
	    return (FAILURE);
	}
	/*
	 * This setbuffer() call was originally used to enhance
	 * the performance of tape_dup by avoiding memory copies.
	 * Unfortunatly, with variable length records read in from
	 * half-inch tapes, our fread()'s don't read into the start
	 * beginning of data_buffer, but rather an offset pointed to
	 * by fp->_ptr.  In any case to get tape_dup working, I've
	 * removed this call and will suffer the overhead of memory
	 * copies.  Why not, isn't that the Unix way anyhow?
	 */
/*	setbuffer (fp, (char *)data_buffer, block_size);	*/
	strcat (bufp, ".info");
	if ( (fpi = fopen (bufp, "r")) == NULL) {
	    fprintf (stderr, "%s: Unable to open info file %s\n",
							name, bufp);
	    return (FAILURE);
	}
	if (debug_flag) {
	    printf ("Processing file number %d...\n", file_number);
	}
#if defined(TIME_TAPE)
	ftime (&btime);
#endif /* defined(TIME_TAPE) */
	return (SUCCESS);
}

/************************************************************************
 *									*
 * tw_proc_eof() - Process disk end of file.				*
 *									*
 * Description:	This function is used to write a file mark to the tape	*
 *		when the disk end of file is encountered.		*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE.				*
 *									*
 ************************************************************************/
#if defined(_QNX_SOURCE)
int
tw_proc_eof(void)
{
	if (verify_flag == FALSE) {
	    QIC02_MSG_STRUCT qic02ms;
	    /*
	     * Write a tape file mark between each tape file.
	     */
	    qic02ms.Header.Command = T_WRITE_FM;
	    if (qnx_ioctl (ofd,
		       QCTL_RAW_CMD,
		       &qic02ms, sizeof (QIC02_HEADER_STRUCT),
		       &qic02ms, sizeof (QIC02_MSG_STRUCT)) == -1) {
		perror ("Write EOF");
		return (FAILURE);
	    }
	}
	if (verbose_flag || debug_flag) {
	    printf ("File %d: Processed %d records containing %d bytes",
				file_number, record_count, file_size);
	    report_time();
	}
	return (SUCCESS);
}
#else /* !defined(_QNX_SOURCE) */
int
tw_proc_eof()
{
	struct mtop op;
	int status = SUCCESS;

	if (verify_flag == FALSE) {
	    /*
	     * Write a tape file mark between each tape file.
	     */
	    op.mt_op = MTWEOF;
	    op.mt_count = (daddr_t) 1;
	    if ((status = ioctl (ofd, MTIOCTOP, (char *)&op)) < 0) {
		perror ("Write EOF");
		return (status);
	    }
	}
	if (verbose_flag || debug_flag) {
	    printf ("File %d: Processed %d records containing %d bytes",
				file_number, record_count, file_size);
	    report_time();
	}
	return (status);
}
#endif /* defined(_QNX_SOURCE) */
